# Start by importing required packages: 
# **pyNN.nest** for the neural simulator
# 
# **Numpy** for efficient use of matrices and arrays
# 
# **Random** for generating random numbers when needed

import pyNN.nest as sim
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from matplotlib import pyplot as plt

# Change "choice" value to get the dataset variant you desire.
#1.1 = OHE-upsampled-unscaled
#1.2 = OHE-upsampled-scaled
#1.3 = OHE-downsampled-unscaled
#1.4 = OHE-downsampled-scaled
#2.1 = IE-upsampled-unscaled
#2.2 = IE-upsampled-scaled
#2.3 = IE-downsampled-unscaled
#2.4 = IE-downsampled-scaled

choice = 1.1
#choice = 1.2

if choice == 1.1:
    data = 'Data/OHE/us/upsampled.csv'
elif choice == 1.2:
    data = 'Data/OHE/us/upsampledScaled.csv'
elif choice == 1.3:
    data = 'Data/OHE/ds/downsampled.csv'
elif choice == 1.4:
    data = 'Data/OHE/ds/downsampledScaled.csv'
elif choice == 2.1:
    data = 'Data/IE/us/upsampled.csv'
elif choice == 2.2:
    data = 'Data/IE/us/upsampledScaled.csv'
elif choice == 2.3:
    data = 'Data/IE/ds/downsampled.csv'
elif choice == 2.4:
    data = 'Data/IE/ds/downsampledScaled.csv'
else:
    print(f'\n"choice" value of {choice} is not valid. Please check it and rerun. Program will now exit...')
    exit() 

# Read the data
s_scores = pd.read_csv(data)
print(s_scores)


# This tells me the number of neurons per input feature (max value in the data +1) and number of neurons in input layer (neurons per input feature * last index in data).
print(s_scores.max(),s_scores.info())


# We will now move on to create the train sources that will be used to train the neural network.

# This method initializes the simulator with update time
def init():
    sim.setup(timestep=TIME_STEP, min_delay=TIME_STEP, max_delay=TIME_STEP, debug=0)


# Then we split the data to train and test (70% - 30%) for purpose of training and evaluating the model respectively.
train, test = train_test_split(s_scores, test_size= 0.3, random_state = 11)

#Quickly inspect the data to make sure everything is good
print(f'lenght of Train: {len(train)}')
print(f'lenght of Test: {len(test)}')
print(train.head(5))
print(test.head(5))
# Looking at the train and test data we can see that the indexes are not serial and this can cause problems during iterations so we reset the index.

#Train index reset
train = train.reset_index()
train.drop(['index'], inplace= True, axis= 1)

#Test index reset
test = test.reset_index()
test.drop(['index'], inplace= True, axis= 1)

#Save the train and test data to a csv file so i can call it in the evaluation notebook and compare
#against the neural nets classifications
train.to_csv(r'output/train.csv', index = False)
test.to_csv(r'output/test.csv', index = False)

TIME_STEP = 1.0
FIRING_WEIGHT = 0.1


# Now i define the number of training epochs and Example time which will determine the number of times the network is trained with the data.
NO_TRAINING_EPOCHS = 4
EXAMPLE_TIME = 30.0

# Here we save the example time and start time so it will just change in the evaluation notebook when running it.
#exampleTime = EXAMPLE_TIME
#testStartTime = END_OF_TRAINING
#%store exampleTime
#%store testStartTime


# Here we calculate the input and output target spike times for the training data which will is used by the spike source array as instruction on when to fire spikes for a corresponding example.
def createTrainSources(data):
    numberDataItems = len(data)

    start = 10.0
    learnOffset = 3
    epochTime = EXAMPLE_TIME*numberDataItems
    spikeSources = []
    inputSpikeTimesMatrix = []
    outputSpikeTimesMatrix = []
    for dataOffset in range (0,numberDataItems):
        #print data[dataOffset]
        inputSpikeTimes = []
        outputSpikeTimes = []

        for epoch in range (0,NO_TRAINING_EPOCHS): 
            inputSpikeTime = start+(dataOffset*EXAMPLE_TIME)+(epoch*epochTime)
            inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
            outputSpikeTime = inputSpikeTime + learnOffset
            outputSpikeTimes = outputSpikeTimes + [outputSpikeTime]
        inputSpikeTimesMatrix = inputSpikeTimesMatrix + [inputSpikeTimes]
        outputSpikeTimesMatrix = outputSpikeTimesMatrix + [outputSpikeTimes]


    inputSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': inputSpikeTimesMatrix},label="inputSpikes")

    outputSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': outputSpikeTimesMatrix},label="outputSpikes")

    return [inputSpikeGens,outputSpikeGens,inputSpikeTimesMatrix]

#Used to check and initialize the end of training
trainSpikeTimes = createTrainSources(train)
END_OF_TRAINING = trainSpikeTimes[2][-1][-1]
print(f'END_OF_TRAINING = {END_OF_TRAINING}')


# Here the input spike times are calculated for the test data and used by another spike source array population.
def createTestSources(data,startTime):
    numberDataItems = len(data)
    timeBetweenTests = EXAMPLE_TIME
    spikeSources = []
    testSpikeTimesMatrix = []
    for dataOffset in range (0,numberDataItems):
        testSpikeTimes = [[startTime + (dataOffset*timeBetweenTests)]]
        testSpikeTimesMatrix = testSpikeTimesMatrix + testSpikeTimes

    testSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': testSpikeTimesMatrix},label="testSpikes")

    return testSpikeGens, testSpikeTimesMatrix

#Used to check the run duration
testSpikeTimes = createTestSources(test, END_OF_TRAINING)
RUN_DURATION = testSpikeTimes[1][-1][-1]
print(f'RUN_DURATION = {RUN_DURATION}')


##The "neuron" and "colIndex" variables are just used to make changing several values more efficient during model evaluation.
#They represent different variables that always have the same value.
key = data.split(sep='/')
key1 = key[2]
key2 = key[3].find('Scaled')
key3 = key[1]

#Initialise "neurons" based on the dataset type
if key1 == 'us' and key2 == -1:
    neurons = 55
elif (key1 == 'us' or key1 == 'ds') and key2 != -1:
    neurons = 6
elif key1 == 'ds' and key2 == -1:
    neurons = 41

#Initialise "colIndex" based on the Encoding type of the used dataset
if key3 == 'OHE':
    colIndex = 41
elif key3 == 'IE':
    colIndex = 32

print(f'neurons:{neurons}\ncolIndex:{colIndex}')
#neurons = upsampled-unscaled:55 | upsampled-scaled:6 |  downsampled-unscaled:41 | downsampled-Scaled:6 (This is regardless of whether it is a One-Hot Encoded or Integer Encoded dataset)
#colIndex =  One-Hot Encoded:41 | Integer Encoded: 32


# Here the connector object is defined which maps which neuron which input and output belongs to.
def convertDataItemToSynapseDescription(dataLine,sourceNumber):
    inputConnector = []
    neuronsPerInputFeature = neurons
    for dataFeature in range (0, colIndex):
        dataFeatureValue = round(dataLine[dataFeature])
        for neuronOffset in range((dataFeatureValue - inputFeatureBreadth),
                              (dataFeatureValue + inputFeatureBreadth+1)):
            if ((neuronOffset >= 0) and (neuronOffset<neuronsPerInputFeature)):
                toNeuron = neuronOffset+(dataFeature*neuronsPerInputFeature)
                inputConnector = inputConnector + [(sourceNumber,toNeuron,
                        FIRING_WEIGHT,TIME_STEP)]

    return inputConnector


# Here the input and output training data is connected through the spike source array to the neurons in the input layer and output layer.
def connectInput(spikeSources, data,inputCells,outputCells):
    inputSpikeSources = spikeSources[0]
    outputSpikeSources = spikeSources[1]
    numberDataItems = len(data)
    for dataOffset in range (0,numberDataItems):
        dataItem = data.values[dataOffset]
        inputConnector=convertDataItemToSynapseDescription(dataItem,dataOffset)
        fromSource = dataOffset
        inputFromListConnector = sim.FromListConnector(inputConnector)
        sim.Projection(inputSpikeSources, inputCells, inputFromListConnector, 
                   receptor_type='excitatory')

        outputCategory = dataItem[colIndex]-1
        outputConnector = [(fromSource,outputCategory,FIRING_WEIGHT,
                            TIME_STEP)]
            
        outputFromListConnector = sim.FromListConnector(outputConnector)
        sim.Projection(outputSpikeSources, outputCells,outputFromListConnector, 
                   receptor_type='excitatory')


# Here the input test data is connected through the other spike source array to the input layer.
def connectTest(testSpikeSources, data,inputCells):
    numberDataItems = len(data)
    for dataOffset in range (0,numberDataItems):
        dataItem = data.values[dataOffset]
        testConnector=convertDataItemToSynapseDescription(dataItem,dataOffset)
        testFromListConnector = sim.FromListConnector(testConnector)
        sim.Projection(testSpikeSources, inputCells, testFromListConnector, 
                   receptor_type='excitatory')


# This method creates the connection between the input and output layer using plastic synapses that are governed by STDP learning rules. Parameters include:
# tau_plus - The time frame before LTP(Long-Term Potentiation) occurs after a neuron fire
# 
# tau_minus - The time frame in which if a neuron fires the weights will be reduced LTD(Long-Term Depression)
# 
# A_plus - How much synaptic weight value increases when LTP occurs
# 
# A_minus - How much the synaptic weight value decreases when LTD occurs
# 
# w_min & w_max - The minimum and maximum synaptic weights possible.
# 
# weight- initial weight
def connectFirstToSecondLayer(fromCells, toCells):
    global firstSecond
    connector = []
    stdp_model = sim.STDPMechanism(
        timing_dependence=sim.SpikePairRule(tau_plus=16.0,tau_minus=16.0,
                                            A_plus=0.003, A_minus=0.014),
        weight_dependence=sim.AdditiveWeightDependence(w_min=0.0, w_max=0.03),
        weight=0.0000,
        delay=1.0,
        dendritic_delay_fraction=float(1))

    firstSecond = sim.Projection(fromCells,toCells,
                   sim.AllToAllConnector(),
                   synapse_type = stdp_model)


# The method is used to create the population of neurons in the input and output layer. The neurons contained are Leaky-Integrate and Fire neurons with exponentially decaying post-synaptic conductance.
def createDefaultNeurons(numberNeurons):
    cells=sim.Population(numberNeurons,sim.IF_cond_exp, sim.IF_cond_exp.default_parameters)
    return cells


# Parameters used for the population:
print(sim.IF_cond_exp.default_parameters)


# Spikes times from the input and output cells are recorded using this.
def setupRecording(cells):
    cells.record({'spikes'})


# This saves the recorded data as csv file.
def printResults(simCells,file):
    simCells.write_data(file+".pkl",'spikes')

#Initialise the simulator
init()

#Create the Spike Source Arrays
trainSpikeGenerators = createTrainSources(train)
testSpikeGenerators = createTestSources(test, END_OF_TRAINING)[0]

#Define the number of neurons in first and second layer and the number of neurons
#stimulated alongside directly stimulated neurons
neuronsInFirstLayer = neurons*colIndex
neuronsInSecondLayer = 4
inputFeatureBreadth = 3

#Create the input and output layer 
firstLayerCells = createDefaultNeurons(neuronsInFirstLayer)
secondLayerCells = createDefaultNeurons(neuronsInSecondLayer)

#Connect the input and output of training data and test data to the spike source arrays and the spike arrays to
#neurons in the input layer and output layer for training.
connectInput(trainSpikeGenerators, train, firstLayerCells, secondLayerCells)
connectTest(testSpikeGenerators, test, firstLayerCells)

#Connect the input layer to the output layer
connectFirstToSecondLayer(firstLayerCells, secondLayerCells)

#record the spike timings from the input and output layer
setupRecording(firstLayerCells)
setupRecording(secondLayerCells)

#Run the simulation
sim.run(RUN_DURATION)

#Get the pickled files of the recordings
printResults(firstLayerCells,"output/first")
printResults(secondLayerCells,"output/second")

#synapseWeights = firstSecond.get(["weight"],format="list")
#print(synapseWeights)

print(neuronsInFirstLayer,firstSecond.size())

print(f'END_OF_TRAINING = {END_OF_TRAINING}')

sim.reset()

print('<<<<<SIMULATION COMPLETE>>>>>')
