# Start by importing required packages:
#
# **pyNN.nest** for the neural simulator
#
# **Numpy** for efficient use of matrices and arrays
#
# **Random** for generating random numbers when needed

import missingno
import pandas as pd
from sklearn.compose import make_column_transformer
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder
from sklearn.utils import resample

# Change "choice" value to generate the dataset variant you want.
# 1.1 = OHE-upsampled-unscaled
# 1.2 = OHE-upsampled-scaled
# 1.3 = OHE-downsampled-unscaled
# 1.4 = OHE-downsampled-scaled
# 2.1 = IE-upsampled-unscaled
# 2.2 = IE-upsampled-scaled
# 2.3 = IE-downsampled-unscaled
# 2.4 = IE-downsampled-scaled

#choice = 1.1
choice = 2.4

s_scores = pd.read_csv('studentscrs.csv')
print(s_scores)


# We check for missing values first
missingno.matrix(s_scores)

# <<<<<<DATA ENCODING>>>>>>>
# All categorical features in the data will be numerically represented with the Integer encoding and one-hot encoding so we can create train sources that will be usable by the system.

enc_choice = str(choice)

if enc_choice.find('1.') != -1:
    # ONE-HOT ENCODING
    ct = make_column_transformer((OneHotEncoder(drop='first'),
                                  ['school', 'sex', 'address', 'famsize',
                                   'Pstatus', 'Mjob', 'Fjob', 'reason',
                                   'guardian', 'schoolsup', 'famsup',
                                 'paid', 'activities', 'nursery', 'higher',
                                   'internet', 'romantic']),
                                 remainder='passthrough')
    odf = pd.DataFrame(ct.fit_transform(s_scores))
    s_scores = odf.astype('int64')
    s_scores.columns = ct.get_feature_names()
    op1 = 'OHE'
elif enc_choice.find('2.') != -1:
    # INTEGER ENCODING
    replace_map = {'school': {'GP': 1, 'MS': 2},
                   'sex': {'M': 1, 'F': 2},
                   'address': {'U': 1, 'R': 2},
                   'famsize': {'LE3': 1, 'GT3': 2},
                   'Pstatus': {'T': 1, 'A': 2},
                   'Mjob': {'teacher': 1, 'health': 2, 'services': 3, 'at_home': 4, 'other': 5},
                   'Fjob': {'teacher': 1, 'health': 2, 'services': 3, 'at_home': 4, 'other': 5},
                   'reason': {'home': 1, 'reputation': 2, 'course': 3, 'other': 4},
                   'guardian': {'father': 1, 'mother': 2, 'other': 3},
                   'schoolsup': {'yes': 1, 'no': 2},
                   'famsup': {'yes': 1, 'no': 2},
                   'paid': {'yes': 1, 'no': 2},
                   'activities': {'yes': 1, 'no': 2},
                   'nursery': {'yes': 1, 'no': 2},
                   'higher': {'yes': 1, 'no': 2},
                   'internet': {'yes': 1, 'no': 2},
                   'romantic': {'yes': 1, 'no': 2}}

    s_scores.replace(replace_map, inplace=True)
    op1 = 'IE'
else:
    print('"choice" is invalid. Please check and rerun')
    exit()

print(s_scores)

# Now that the categorical features have been encoded with their respective numerical values, G3 will now be addressed seperately to create the classification target which will be done by creating bins for the scores and assigning each bin with the corresponding label:
# G3(becomes classification target) --- 0-5(Fail) - 1, 6-10(Pass) - 2, 11-15(Credit) - 3, 16-20(Distinction) - 4
bin_labels = [1, 2, 3, 4]
s_scores['target'] = pd.cut(s_scores['G3'], bins=4, labels=bin_labels)
s_scores['target'] = s_scores['target'].astype('int64')
print(s_scores.info())

# G3 will now be dropped as its no longer needed.
s_scores.drop('G3', inplace=True, axis=1)


# <<<<<FEATURE SCALING>>>>>
# Here i scale all the continuous features in the data.
if choice == 1.2 or choice == 1.4 or choice == 2.2 or choice == 2.4:
    conCols = ['age', 'traveltime', 'studytime',
               'absences', 'failures', 'G1', 'G2']
    scaler = MinMaxScaler()
    op3 = 'Scaled'
    print('Scaled:\n', s_scores)
    s_scores[conCols] = scaler.fit_transform(s_scores[conCols])
elif choice == 1.1 or choice == 1.3 or choice == 2.1 or choice == 2.3:
    print('Unscaled:\n', s_scores)
    op3 = 'Unscaled'
else:
    print('"choice" is invalid. Please check and rerun')
    exit()

# Now for class balancing. Lets check the target class ratios
print(s_scores['target'].value_counts())

# <<<<<<CLASS BALANCING>>>>>>
# Seperate the data by class
class_one = s_scores[s_scores.target == 1]
class_two = s_scores[s_scores.target == 2]
class_three = s_scores[s_scores.target == 3]
class_four = s_scores[s_scores.target == 4]

u_sample = 169
d_sample = 40

if choice == 1.1 or choice == 1.2 or choice == 2.1 or choice == 2.2:
    # Upsampling
    class_one_rs = resample(class_one, replace=True,
                            n_samples=u_sample, random_state=11)
    class_two_rs = resample(class_two, replace=True,
                            n_samples=u_sample, random_state=11)
    class_four_rs = resample(class_four, replace=True,
                             n_samples=u_sample, random_state=11)
    finalScores = pd.concat(
        [class_one_rs, class_two_rs, class_three, class_four_rs])
    op2 = 'Upsampled'
elif choice == 1.3 or choice == 1.4 or choice == 2.3 or choice == 2.4:
    # downsampling
    class_one_rs = resample(class_one, replace=True,
                            n_samples=d_sample, random_state=11)
    class_two_rs = resample(class_two, replace=True,
                            n_samples=d_sample, random_state=11)
    class_three_rs = resample(
        class_three, replace=True, n_samples=d_sample, random_state=11)
    finalScores = pd.concat(
        [class_one_rs, class_two_rs, class_three_rs, class_four])
    op2 = 'Downsampled'
else:
    print('"choice" is invalid. Please check and rerun')
    exit()

print(finalScores['target'].value_counts())

# <<<<<DATA EXPORT TO CSV>>>>>
if choice == 1.1:
    data = 'Data/OHE/us/upsampled.csv'
elif choice == 1.2:
    data = 'Data/OHE/us/upsampledScaled.csv'
elif choice == 1.3:
    data = 'Data/OHE/ds/downsampled.csv'
elif choice == 1.4:
    data = 'Data/OHE/ds/downsampledScaled.csv'
elif choice == 2.1:
    data = 'Data/IE/us/upsampled.csv'
elif choice == 2.2:
    data = 'Data/IE/us/upsampledScaled.csv'
elif choice == 2.3:
    data = 'Data/IE/ds/downsampled.csv'
elif choice == 2.4:
    data = 'Data/IE/ds/downsampledScaled.csv'

try:
    finalScores.to_csv(data, index=False)
    print('Data Variant generated: ', op1 + '-' + op2 + '-' + op3)
except FileNotFoundError:
    print('<<<ERROR>>>\nCSV not created.\nPlease create the required folders:\n' +
          '1. Data/OHE/us/\n2. Data/OHE/ds\n3. Data/IE/us\n4. Data/IE/ds')
